// Ultrastik.cs
//
// This program is free software. It is dual licensed under GNU GPL and GNU LGPL.
// See COPYING_GPL.txt and COPYING_LGPL.txt for details.
//

#region Ultimarc Ultrastik Info
// Ultimarc Ultrastik Class
//
// The Ultrastik 360 is a full-speed USB 2.0 composite device.  Communication
// to the Ultrastik is done through Vendor Request Control Transfers.
//
// Notes:  
// 1. The Ultrastik is a full-speed USB 2.0 (not high speed).
// 2. Firware upgrade 1.7 to 1.9(Beta) or higher changes the device descriptors.  In 
//    Windows this is not recommended and the user should delete or uninstall all instances
//    of the Ultrastik for proper operation.
//
// Ulimarc Ultrastik VID = 0xD209
// Ulimarc Ultrastik VID = 0x0501, 0x0502, 0x0503, 0x0504
//
// Communication
//
// Control Transfer Types (bmRequesType, bRequest, wValue, Data,wLength)
// "Start Message" (0x43,0xE9,0x0001,No Data,0)
// "Data Write" (0x43,0xEB,0x0000,Data,(Size of Data))
// "Data Confirm" (0xC3,0xEA,0x0000,No Data,1) (returns 1 byte)
// "End Message" (0x43,0xE9,0x0000,No Data,0)
// 
// ************ General Transaction Format*************
// All Transactions follow this format.  The only difference
// is the # of "Data" Transfers and Confirms
//
// "Start Message" Transfer
// "Data Write" Transfer, MAX size of 32 Bytes per Transfer
// "Data Confirm" Transfer, follows each "Data Write" Transfer
//  repeat "Data Write" and "Data Confirm" for all Data.
// "End Message" Transfer
//
// ********************Change Ultrastik ID Transfer*********************
// 1 "Data Write" Transfer of 32 Bytes
// First Byte determines Ultrastik ID
// 0x51 = Ultrastik Device #1
// 0x52 = Ultrastik Device #2
// 0x53 = Ultrastik Device #3
// 0x54 = Ultrastik Device #4
//
// *********************Send Map Transfer*******************************
// This is only for Map Size = 9
//
// 3 "Data Write" Transfers of 32 Bytes
// -First Transfer-
// Byte[0] = 0x50 (Header)
// Byte[1] = 0x09 (Map Size?)
// Byte[2] = Restrictor (RestrictorOn = 0x09, RestrictorOff = 0x10)
// Byte[3]-Byte[10] = Map Border Location
// Byte[11]-Byte[31] = Mapping of Block Location. starting from topleft and read left to right.
//
// -Second Transfer-
// 32 Bytes - Continue Mapping of Block Location
//
// -Third Transfer-
// Byte[0]-Byte[27] - Continue Mapping of Block Location
// Byte[28]-Byte[30] - 0x00
// Byte[31] - (Write to Flash 0x00) (Write to RAM 0xFF, supported in Firmware 2.2)
//
//
// ****************************************************
// USB      Mapping
// 0x00 =   Analog (-)
// 0x01 =   Center (C)
// 0x02 =   Up (N)
// 0x03 =   UPRight (NE)
// 0x04 =   Right (E)
// 0x05 =   DownRight (SE)
// 0x06 =   Down (S)
// 0x07 =   DownLeft (SW)
// 0x08 =   Left (W)
// 0x09 =   UPLeft (NW)
// 0x0A =   Sticky (*)
#endregion

using System;
using System.IO;
using System.Collections.Generic;
using System.Text;
using ICSharpCode.USBlib;

namespace Ultimarc
{
    ///<summary>
    ///Ultimarc Ultrastik360 Class
    ///</summary>
    class Ultrastik : IDisposable, IComparable
    {

        const byte RestrictorOn = 0x09;
        const byte RestrictorOff = 0x10;
        const byte Flash = 0x00;
        const byte RAM = 0xFF;
        byte[] MapBorderLocation = {0x1E,0x3A,0x56,0x72,0x8E,0xAA,0xC6,0xE2};

	    Device device1;

        byte restrictor;
        byte flash;


        #region vjoy2way_buffers
        static byte[] buffer1_vjoy2way =    {0x50,0x09,0x10,0x1E,0x3A,0x56,0x72,0x8E,
                                           0xAA,0xC6,0xE2,0x02,0x02,0x02,0x02,0x02,
                                           0x02,0x02,0x02,0x02,0x02,0x02,0x02,0x02,
                                           0x02,0x02,0x02,0x02,0x02,0x02,0x02,0x02};

        static byte[] buffer2_vjoy2way =    {0x02,0x02,0x02,0x02,0x02,0x02,0x01,0x01,
                                           0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,
                                           0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,
                                           0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01};

        static byte[] buffer3_vjoy2way =    {0x01,0x06,0x06,0x06,0x06,0x06,0x06,0x06,
                                           0x06,0x06,0x06,0x06,0x06,0x06,0x06,0x06,
                                           0x06,0x06,0x06,0x06,0x06,0x06,0x06,0x06,
                                           0x06,0x06,0x06,0x06,0x00,0x00,0x00,0x00};
        #endregion

        #region joy2way_buffers
        static byte[] buffer1_joy2way =    {0x50,0x09,0x10,0x1E,0x3A,0x56,0x72,0x8E,
                                           0xAA,0xC6,0xE2,0x08,0x08,0x08,0x01,0x01,
                                           0x01,0x04,0x04,0x04,0x08,0x08,0x08,0x01,
                                           0x01,0x01,0x04,0x04,0x04,0x08,0x08,0x08};

        static byte[] buffer2_joy2way =    {0x01,0x01,0x01,0x04,0x04,0x04,0x08,0x08,
                                           0x08,0x01,0x01,0x01,0x04,0x04,0x04,0x08,
                                           0x08,0x08,0x01,0x01,0x01,0x04,0x04,0x04,
                                           0x08,0x08,0x08,0x01,0x01,0x01,0x04,0x04};

        static byte[] buffer3_joy2way =    {0x04,0x08,0x08,0x08,0x01,0x01,0x01,0x04,
                                           0x04,0x04,0x08,0x08,0x08,0x01,0x01,0x01,
                                           0x04,0x04,0x04,0x08,0x08,0x08,0x01,0x01,
                                           0x01,0x04,0x04,0x04,0x00,0x00,0x00,0x00};
        #endregion

        #region joy4way_buffers
        static byte[] buffer1_joy4way =    {0x50,0x09,0x10,0x1E,0x3A,0x56,0x72,0x8E,
                                           0xAA,0xC6,0xE2,0x0A,0x02,0x02,0x02,0x02,
                                           0x02,0x02,0x02,0x0A,0x08,0x0A,0x02,0x02,
                                           0x02,0x02,0x02,0x0A,0x04,0x08,0x08,0x0A};

        static byte[] buffer2_joy4way =    {0x02,0x02,0x02,0x0A,0x04,0x04,0x08,0x08,
                                           0x08,0x01,0x01,0x01,0x04,0x04,0x04,0x08,
                                           0x08,0x08,0x01,0x01,0x01,0x04,0x04,0x04,
                                           0x08,0x08,0x08,0x01,0x01,0x01,0x04,0x04};

        static byte[] buffer3_joy4way =    {0x04,0x08,0x08,0x0A,0x06,0x06,0x06,0x0A,
                                           0x04,0x04,0x08,0x0A,0x06,0x06,0x06,0x06,
                                           0x06,0x0A,0x04,0x0A,0x06,0x06,0x06,0x06,
                                           0x06,0x06,0x06,0x0A,0x00,0x00,0x00,0x00};
        #endregion

        #region diagonaljoy4way_buffers
        static byte[] buffer1_diagonaljoy4way =    {0x50,0x09,0x10,0x1E,0x3A,0x56,0x72,0x8E,
                                           0xAA,0xC6,0xE2,0x09,0x09,0x09,0x09,0x0A,
                                           0x03,0x03,0x03,0x03,0x09,0x09,0x09,0x09,
                                           0x0A,0x03,0x03,0x03,0x03,0x09,0x09,0x09};

        static byte[] buffer2_diagonaljoy4way =    {0x09,0x01,0x03,0x03,0x03,0x03,0x09,0x09,
                                           0x09,0x01,0x01,0x01,0x03,0x03,0x03,0x0A,
                                           0x0A,0x01,0x01,0x01,0x01,0x01,0x0A,0x0A,
                                           0x07,0x07,0x07,0x01,0x01,0x01,0x05,0x05};

        static byte[] buffer3_diagonaljoy4way =    {0x05,0x07,0x07,0x07,0x07,0x01,0x05,0x05,
                                           0x05,0x05,0x07,0x07,0x07,0x07,0x0A,0x05,
                                           0x05,0x05,0x05,0x07,0x07,0x07,0x07,0x0A,
                                           0x05,0x05,0x05,0x05,0x00,0x00,0x00,0x00};
        #endregion

        #region joy8way_buffers
        static byte[] buffer1_joy8way =    {0x50,0x09,0x10,0x1E,0x3A,0x56,0x72,0x8E,
                                           0xAA,0xC6,0xE2,0x09,0x09,0x09,0x02,0x02,
                                           0x02,0x03,0x03,0x03,0x09,0x09,0x09,0x02,
                                           0x02,0x02,0x03,0x03,0x03,0x09,0x09,0x09};

        static byte[] buffer2_joy8way =    {0x02,0x02,0x02,0x03,0x03,0x03,0x08,0x08,
                                           0x08,0x01,0x01,0x01,0x04,0x04,0x04,0x08,
                                           0x08,0x08,0x01,0x01,0x01,0x04,0x04,0x04,
                                           0x08,0x08,0x08,0x01,0x01,0x01,0x04,0x04};

        static byte[] buffer3_joy8way =    {0x04,0x07,0x07,0x07,0x06,0x06,0x06,0x05,
                                           0x05,0x05,0x07,0x07,0x07,0x06,0x06,0x06,
                                           0x05,0x05,0x05,0x07,0x07,0x07,0x06,0x06,
                                           0x06,0x05,0x05,0x05,0x00,0x00,0x00,0x00};
        #endregion

        #region joyeasy8way_buffers
        static byte[] buffer1_joyeasy8way =    {0x50,0x09,0x10,0x1E,0x3A,0x56,0x72,0x8E,
                                           0xAA,0xC6,0xE2,0x09,0x09,0x0A,0x02,0x02,
                                           0x02,0x0A,0x03,0x03,0x09,0x09,0x09,0x02,
                                           0x02,0x02,0x03,0x03,0x03,0x0A,0x09,0x09};

        static byte[] buffer2_joyeasy8way =    {0x0A,0x02,0x0A,0x03,0x03,0x0A,0x08,0x08,
                                           0x0A,0x09,0x01,0x03,0x0A,0x04,0x04,0x08,
                                           0x08,0x08,0x01,0x01,0x01,0x04,0x04,0x04,
                                           0x08,0x08,0x0A,0x07,0x01,0x05,0x0A,0x04};

        static byte[] buffer3_joyeasy8way =    {0x04,0x0A,0x07,0x07,0x0A,0x06,0x0A,0x05,
                                           0x05,0x0A,0x07,0x07,0x07,0x06,0x06,0x06,
                                           0x05,0x05,0x05,0x07,0x07,0x0A,0x06,0x06,
                                           0x06,0x0A,0x05,0x05,0x00,0x00,0x00,0x00};
        #endregion

        #region analog_buffers
        static byte[] buffer1_analog =    {0x50,0x09,0x10,0x1E,0x3A,0x56,0x72,0x8E,
                                           0xAA,0xC6,0xE2,0x00,0x00,0x00,0x00,0x00,
                                           0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
                                           0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00};

        static byte[] buffer2_analog =    {0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
                                           0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
                                           0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
                                           0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00};
        #endregion

        #region mouse_buffers
        static byte[] buffer1_mouse =    {0x50,0x09,0x10,0x1E,0x3A,0x56,0x72,0x8E,
                                           0xAA,0xC6,0xE2,0x01,0x01,0x01,0x01,0x01,
                                           0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,
                                           0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01};

        static byte[] buffer2_mouse =    {0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,
                                           0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,
                                           0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,
                                           0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01};

        static byte[] buffer3_mouse =    {0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,
                                           0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,
                                           0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,
                                           0x01,0x01,0x01,0x01,0x00,0x00,0x00,0x00};

        #endregion


        Ultrastik(Device device)
        {
            this.device1 = device;
            this.Restrictor = false;
            this.UseFlash = false;
        }

        /// <summary>
        /// Returns the USB Vendor ID(VID) of the Ultrastik
        /// </summary>
        /// <exception cref="None"></exception>
        public int VendorID
        {
            get
            {
                return this.device1.Descriptor.VendorId;
            }
        }


        /// <summary>
        /// Returns the USB Product ID(PID) of the Ultrastik
        /// </summary>
        /// <exception cref="None"></exception>
        public int ProductID
        {
            get
            {
                return this.device1.Descriptor.ProductId;
            }
        }


        /// <summary>
        /// Returns the USB Manufacturer String of the Ultrastik
        /// </summary>
        /// <exception cref="None"></exception>
        public string Manufacturer
        {
            get
            {
                return this.device1.Manufacturer;
            }
        }


        /// <summary>
        /// Returns the USB Product String of the Ultrastik
        /// </summary>
        /// <exception cref="None"></exception>
        public string Product
        {
            get
            {
                return this.device1.Product;
            }
        }


        /// <summary>
        /// Returns the USB Serial Number of the Ultrastik
        /// </summary>
        /// <exception cref="None"></exception>
        public string SerialNumber
        {
            get
            {
                return this.device1.SerialNumber;
            }
        }


        /// <summary>
        /// Returns the Firmware Version of the Ultrastik
        /// </summary>
        /// <exception cref="None"></exception>
        public int FirmwareVersion
        {
            get
            {
                return this.device1.Descriptor.bcdDevice;
            }
        }


        /// <summary>
        /// Opens the devices into an array.
        /// </summary>
        /// <exception cref="ICSharpCode.USBlib.UsbException"></exception>
        /// <returns>An Array of Ultrastik Devices</returns>
        public static Ultrastik[] Open()
        {
            List<Ultrastik> list1 = new List<Ultrastik>();        

            try
            {             
                foreach (Bus bus in Bus.Busses)
                {
                    foreach (Descriptor descriptor in bus.Descriptors)
                    {
                        if (descriptor.VendorId == 0xD209 && (descriptor.ProductId >= 0x0501 && descriptor.ProductId <= 0x0504))
                        {
                                list1.Add(new Ultrastik(descriptor.OpenDevice()));
                        }
                    }
                }
            }
            catch (UsbException e)
            {
                Console.WriteLine("Got Exception : " + e);
            }

            return (Ultrastik[])list1.ToArray();
        }

        /// <summary>
        /// Opens the devices into an array.
        /// </summary>
        /// <exception cref="ICSharpCode.USBlib.UsbException"></exception>
        /// <param name="VENDOR_ID">Hexadecimal Vendor ID for the device</param>
        /// <param name="PRODUCT_ID">Hexadecimal Product ID for the device</param>
        /// <returns>An Array of Ultrastik Devices</returns>
        public static Ultrastik[] Open(int VENDOR_ID, int PRODUCT_ID)
        {
            List<Ultrastik> list1 = new List<Ultrastik>();

            try
            {
                foreach (Bus bus in Bus.Busses)
                {
                    foreach (Descriptor descriptor in bus.Descriptors)
                    {
                        if (descriptor.VendorId == VENDOR_ID && descriptor.ProductId == PRODUCT_ID)
                        {
                            list1.Add(new Ultrastik(descriptor.OpenDevice()));
                        }
                    }
                }
            }
            catch (UsbException e)
            {
                Console.WriteLine("Got Exception : " + e);
            }

            return (Ultrastik[])list1.ToArray();
        }


        /// <summary>
        /// Sets the Restrictor for the current device.
        /// Restrictor is programmed when sending a Map.
        /// </summary>
        /// <exception cref="None"></exception>
        public bool Restrictor
        {
            set
            {
                if (value)
                    this.restrictor = RestrictorOn;
                else
                    this.restrictor = RestrictorOff;                
            }
        }


        /// <summary>
        /// Determines if the MAP is written to Flash or RAM.
        /// Programmed when sending a Map.
        /// </summary>
        /// <exception cref="None"></exception>
        public bool UseFlash
        {
            set
            {
                if (value)
                    this.flash = Flash;
                else
                    this.flash = RAM;
            }
        }


        /// <summary>
        /// Gets/Sets the Joystick ID for the current device.
        /// User must unplug and reattach device afterwards.
        /// </summary>
        /// <exception cref="ICSharpCode.USBlib.UsbException"></exception>
        public int UltrastikID
        {
            get
            {
                return this.device1.Descriptor.ProductId;
            }
            set
            {
                byte[] tempbuffer = new byte[31];//buffer_ID;

                switch (value)
                {
                    case 1:
                        {
                            tempbuffer[0] = 0x51;
                            break;
                        }
                    case 2:
                        {
                            tempbuffer[0] = 0x52;
                            break;
                        }
                    case 3:
                        {
                            tempbuffer[0] = 0x53;
                            break;
                        }
                    case 4:
                        {
                            tempbuffer[0] = 0x54;
                            break;
                        }
                    default:
                        {   
                            return;
                        }
                }
                try
                {
                    this.device1.SendControlMessage(0x43, 0xE9, 0x0001, null);  //Start Message Transfer

                    this.device1.SendControlMessage(0x43, 0xEB, 0x0000, tempbuffer);   //Data Write Transfer
                    this.device1.SendControlMessage(0xC3, 0xEA, 0x0000, null, 1);   //Data Confirm Transfer

                    this.device1.SendControlMessage(0x43, 0xE9, 0x0000, null);  //End Message Transfer
                }
                catch (UsbException e)
                {
                    Console.WriteLine("Got Exception : " + e);
                }

            }
        }


        /// <summary>
        /// Loads Hardcoded Maps to the Current device.
        /// </summary>
        /// <exception cref="ICSharpCode.USBlib.UsbException"></exception>
        public bool LoadMap(string map)
        {
            byte[] tempbuffer1 = new byte[31];
            byte[] tempbuffer2 = new byte[31];
            byte[] tempbuffer3 = new byte[31];

            switch (map.ToLower())
            {
                case "vjoy2way":
                    {
                        tempbuffer1 = buffer1_vjoy2way;
                        tempbuffer2 = buffer2_vjoy2way;
                        tempbuffer3 = buffer3_vjoy2way;
                        break;
                    }
                case "joy2way":
                    {
                        tempbuffer1 = buffer1_joy2way;
                        tempbuffer2 = buffer2_joy2way;
                        tempbuffer3 = buffer3_joy2way;
                        break;
                    }
                case "joy4way":
                    {
                        tempbuffer1 = buffer1_joy4way;
                        tempbuffer2 = buffer2_joy4way;
                        tempbuffer3 = buffer3_joy4way;
                        break;
                    }
                case "diagonaljoy4way":
                    {
                        tempbuffer1 = buffer1_diagonaljoy4way;
                        tempbuffer2 = buffer2_diagonaljoy4way;
                        tempbuffer3 = buffer3_diagonaljoy4way;
                        break;
                    }
                case "joy8way":
                    {
                        tempbuffer1 = buffer1_joy8way;
                        tempbuffer2 = buffer2_joy8way;
                        tempbuffer3 = buffer3_joy8way;
                        break;
                    }
                case "joy8wayeasy":
                    {
                        tempbuffer1 = buffer1_joyeasy8way;
                        tempbuffer2 = buffer2_joyeasy8way;
                        tempbuffer3 = buffer3_joyeasy8way;
                        break;
                    }
                case "analog":
                    {
                        tempbuffer1 = buffer1_analog;
                        tempbuffer2 = buffer2_analog;
                        tempbuffer3 = buffer2_analog;
                        break;
                    }
                case "mouse":
                    {
                        tempbuffer1 = buffer1_mouse;
                        tempbuffer2 = buffer2_mouse;
                        tempbuffer3 = buffer3_mouse;
                        break;
                    }
                default: //Else Map Analog
                    {
                        return false;
                    }
            }

            SendMap(tempbuffer1, tempbuffer2, tempbuffer3);

            return true;
        }


        /// <summary>
        /// Loads Custom Map to the Current device.
        /// </summary>
        /// <exception cref="ICSharpCode.USBlib.UsbException"></exception>
        public bool LoadMapFile(string inputfilename)
        {
            string input = null;
            string[] parse;
            byte[] mapbuffer1 = buffer1_mouse;
            byte[] mapbuffer2 = buffer2_mouse;
            byte[] mapbuffer3 = buffer3_mouse;

            try
            {
                if (!File.Exists(inputfilename))
                {
                    return false;
                }
                using (StreamReader sr = File.OpenText(inputfilename.ToString()))
                {
                    while ((input = sr.ReadLine()) != null)
                    {
                        parse = input.Split('=');

                        if (parse.Length == 2)
                        {
                            switch (parse[0])
                            {
                                case "MapFileFormatVersion":
                                    {
                                        if (!parse[1].Equals("1.0"))
                                        {
                                            sr.ReadToEnd();
                                            return false;
                                        }
                                        break;
                                    }
                                case "MapSize":
                                    {
                                        if (!parse[1].Equals("9"))
                                        {
                                            sr.ReadToEnd();
                                            return false;
                                        }
                                        break;
                                    }
                                case "MapBorderLocations":
                                    {
                                        string[] map = parse[1].Split(',');
                                        if (map.Length == 8)
                                        {
                                            for (int i = 0; (i < map.Length); i++)
                                            {
                                                this.MapBorderLocation[i] = Convert.ToByte(map[i]);
                                            }
                                            break;
                                        }
                                        else
                                        {
                                            sr.ReadToEnd();
                                            return false;
                                        }
                                    }
                                case "MapRow1":
                                    {
                                        byte[] row1 = Str2Byte(parse[1].Split(','));
                                        for (int counter = 0; (counter <= 8); counter++)
                                        {
                                            mapbuffer1[counter + 11] = row1[counter];
                                        }
                                        break;
                                    }
                                case "MapRow2":
                                    {
                                        byte[] row2 = Str2Byte(parse[1].Split(','));
                                        for (int counter = 0; (counter <= 8); counter++)
                                        {
                                            mapbuffer1[counter + 20] = row2[counter];
                                        }
                                        break;
                                    }
                                case "MapRow3":
                                    {
                                        byte[] row3 = Str2Byte(parse[1].Split(','));
                                        for (int counter = 0; (counter <= 2); counter++)
                                        {
                                            mapbuffer1[counter + 29] = row3[counter];
                                        }
                                        for (int counter = 0; (counter <= 5); counter++)
                                        {
                                            mapbuffer2[counter] = row3[counter + 3];
                                        }
                                        break;
                                    }
                                case "MapRow4":
                                    {
                                        byte[] row4 = Str2Byte(parse[1].Split(','));
                                        for (int counter = 0; (counter <= 8); counter++)
                                        {
                                            mapbuffer2[counter + 6] = row4[counter];
                                        }
                                        break;
                                    }
                                case "MapRow5":
                                    {
                                        byte[] row5 = Str2Byte(parse[1].Split(','));
                                        for (int counter = 0; (counter <= 8); counter++)
                                        {
                                            mapbuffer2[counter + 15] = row5[counter];
                                        }
                                        break;
                                    }
                                case "MapRow6":
                                    {
                                        byte[] row6 = Str2Byte(parse[1].Split(','));
                                        for (int counter = 0; (counter <= 7); counter++)
                                        {
                                            mapbuffer2[counter + 24] = row6[counter];
                                        }
                                        mapbuffer3[0] = row6[8];
                                        break;
                                    }
                                case "MapRow7":
                                    {
                                        byte[] row7 = Str2Byte(parse[1].Split(','));
                                        for (int counter = 0; (counter <= 8); counter++)
                                        {
                                            mapbuffer3[counter + 1] = row7[counter];
                                        }
                                        break;
                                    }
                                case "MapRow8":
                                    {
                                        byte[] row8 = Str2Byte(parse[1].Split(','));
                                        for (int counter = 0; (counter <= 8); counter++)
                                        {
                                            mapbuffer3[counter + 10] = row8[counter];
                                        }
                                        break;
                                    }
                                case "MapRow9":
                                    {
                                        byte[] row9 = Str2Byte(parse[1].Split(','));
                                        for (int counter = 0; (counter <= 8); counter++)
                                        {
                                            mapbuffer3[counter + 19] = row9[counter];
                                        }
                                        break;
                                    }
                                default:
                                    {
                                        break;
                                    }
                            }
                        }
                    }
                    sr.Close();
                }
            }
            catch (Exception)
            {
                // Let the user know what went wrong.
                Console.WriteLine("Error Loading Map");
            }

            SendMap(mapbuffer1, mapbuffer2, mapbuffer3);

            return true;
        }


        /// <summary>
        /// Converts MAP String[] to a Byte[]
        /// </summary>
        /// <exception cref="None"></exception>
        private static byte[] Str2Byte(string[] mapRow)
        {
            byte[] row = new byte[mapRow.Length];

            for (int i = 0; (i < mapRow.Length); i++)
            {
                switch (mapRow[i])
                {
                    case "-":
                        {
                            row[i] = 0x00;
                            break;
                        }
                    case "C":
                        {
                            row[i] = 0x01;
                            break;
                        }
                    case "N":
                        {
                            row[i] = 0x02;
                            break;
                        }
                    case "NE":
                        {
                            row[i] = 0x03;
                            break;
                        }
                    case "E":
                        {
                            row[i] = 0x04;
                            break;
                        }
                    case "SE":
                        {
                            row[i] = 0x05;
                            break;
                        }
                    case "S":
                        {
                            row[i] = 0x06;
                            break;
                        }
                    case "SW":
                        {
                            row[i] = 0x07;
                            break;
                        }
                    case "W":
                        {
                            row[i] = 0x08;
                            break;
                        }
                    case "NW":
                        {
                            row[i] = 0x09;
                            break;
                        }
                    case "*":
                        {
                            row[i] = 0x0A;
                            break;
                        }
                    default:
                        {
                            break;
                        }
                }
            }

            return row;
        }


        /// <summary>
        /// Sends Map to the Current device.
        /// </summary>
        /// <exception cref="ICSharpCode.USBlib.UsbException"></exception>
        private void SendMap(Byte[] buffer1, Byte[] buffer2, Byte[] buffer3)
        {

            buffer1[2] = this.restrictor;

            if (this.FirmwareVersion >= 0x0202)
            {
                buffer3[31] = this.flash;
                for (int i = 0; i < 7; i++)
                {
                    buffer1[i + 3] = this.MapBorderLocation[i];
                }
            }

            try
            {

                this.device1.SendControlMessage(0x43, 0xE9, 0x0001, null);  //Start Message Transfer

                this.device1.SendControlMessage(0x43, 0xEB, 0x0000, buffer1);   //Data Write Transfer
                this.device1.SendControlMessage(0xC3, 0xEA, 0x0000, null, 1);   //Data Confirm Transfer
 
                this.device1.SendControlMessage(0x43, 0xEB, 0x0000, buffer2);   //Data Write Transfer
                this.device1.SendControlMessage(0xC3, 0xEA, 0x0000, null, 1);   //Data Confirm Transfer

                this.device1.SendControlMessage(0x43, 0xEB, 0x0000, buffer3);   //Data Write Transfer
                this.device1.SendControlMessage(0xC3, 0xEA, 0x0000, null, 1);   //Data Confirm Transfer

                this.device1.SendControlMessage(0x43, 0xE9, 0x0000, null);  //End Message Transfer
            }
            catch (UsbException e)
            {
                Console.WriteLine("Got Exception : " + e);
            }

        }


        //ToDo
        //public void UpdateFirmware()
        //{
        //}

        
        public void Close()
		{
            Dispose();
		}
		
		#region System.IDisposable interface implementation
		public void Dispose() 
		{
            this.device1.Close();
            GC.SuppressFinalize(this);
		}

		~Ultrastik()
		{
			Dispose();
        }
        #endregion

        #region System.IComparable interface implementation
        public int CompareTo(object obj)
        {
            if (this.GetType() != obj.GetType())
            {
                throw (new ArgumentException("Both objects being compared must be of the same type"));
            }
            else
            {
                Ultrastik u360 = (Ultrastik)obj;

                if (this.ProductID == u360.ProductID)
                {
                    return (0);
                }
                else if (this.ProductID > u360.ProductID)
                {
                    return(1);
                }
                else
                {
                    return (-1);
                }
            }
        }
        #endregion

    }
}
